import os
import uvicore
from uvicore.support import str
from uvicore.support.dumper import dump, dd
from uvicore.support.schematic import Schematic
from uvicore.console import command, argument, option


@command()
@argument('name')
@argument('table')
async def model(name: str, table: str):
    """
    \b
    Generate a new ORM model schematic...
    \b
    USAGE:
        Models should be lower_understore and SINGULAR
        While their table should be lower_understore and PLURAL
    \b
        ./uvicore gen model user users
        ./uvicore gen model user_detail user_details
        ./uvicore gen model post posts
        ./uvicore gen model post_tag post_tags
    """

    # Get calling package (main running app)
    package = uvicore.app.package(main=True)

    # Get stub (src)
    stub = os.path.dirname(__file__) + '/stubs/model.py'

    # Get destination for this filetype, considering the packages path customizations
    dest = package.folder_path('models') + '/' + name + '.py'

    Schematic(
        type='model',
        stub=stub,
        dest=dest,
        replace = [
            ('xx_modelname', name),
            ('xx_ModelName', str.studly(name)),
            ('xx_tablename', table),
            ('xx_TableName', str.studly(table)),
        ]
    ).generate()

    uvicore.log.nl()
    uvicore.log.notice('Be sure to add this model to your ./models/__init__.py')



from .group_roles import GroupRoles
from .groups import Groups
from .permissions import Permissions
from .roles import Roles
from .role_permissions import RolePermissions
from .user_groups import UserGroups
from .user_roles import UserRoles
from .users import Users



from abc import ABC, abstractmethod

# class Event(ABC):

#     is_async: bool

# #from uvicore.typing import Dict
# class Event(Dict):
#     """Event Definition"""
#     # These class level properties for for type annotations only.
#     # They do not restrict of define valid properties like a dataclass would.
#     # This is still a fully dynamic SuperDict!
#     name: str
#     description: str
#     dynamic: bool
#     is_async: bool







from .package import Package
from .connection import Connection
from abc import ABC, abstractmethod
from uvicore.contracts import DbQueryBuilder
from uvicore.typing import Any, Dict, List, Sequence, Mapping, Optional

# Optional imports based on installed modules
try:
    import sqlalchemy as sa
except ImportError:
    pass


class Database(ABC):

    @property
    @abstractmethod
    def default(self) -> str:
        """The default connection str for the main running app"""
        pass

    @property
    @abstractmethod
    def connections(self) -> Dict[str, Connection]:
        """All connections from all packages, keyed by connection str name"""
        pass

    @property
    @abstractmethod
    def engines(self) -> Dict[str, sa.engine.Engine]:
        """All engines for all unique (by metakey) connections, keyed by metakey"""
        pass

    @property
    @abstractmethod
    def metadatas(self) -> Dict[str, sa.MetaData]:
        """All SQLAlchemy Metadata for all unique (by metakey) connections, keyed by metakey"""
        pass

    @abstractmethod
    def init(self, default: str, connections: Dict[str, Connection]) -> None:
        """Initialize the database system with a default connection str and List of all Connections from all packages"""
        pass

    @abstractmethod
    def packages(self, connection: str = None, metakey: str = None) -> List[Package]:
        """Get all packages with the metakey (direct or derived from connection str)."""
        pass

    @abstractmethod
    def metakey(self, connection: str = None, metakey: str = None) -> str:
        """Get one metekay by connection str or metakey"""
        pass

    @abstractmethod
    def connection(self, connection: str = None) -> Connection:
        """Get one connection by connection name"""
        pass

    @abstractmethod
    def metadata(self, connection: str = None, metakey: str = None) -> sa.MetaData:
        """Get one SQLAlchemy Metadata by connection str or metakey"""
        pass

    @abstractmethod
    def tables(self, connection: str = None, metakey: str = None) -> List[sa.Table]:
        """Get all SQLAlchemy tables for a given connection str or metakey"""
        pass

    @abstractmethod
    def table(self, table: str, connection: str = None) -> sa.Table:
        """Get one SQLAlchemy Table by name (without prefix) and connection str or connection.tablename dot notation"""
        pass

    @abstractmethod
    def tablename(self, table: str, connection: str = None) -> str:
        """Get a SQLAlchemy tablename with prefix by name (without prefix) and connection str or connection.tablename dot notation"""
        pass

    @abstractmethod
    def engine(self, connection: str = None, metakey: str = None) -> sa.engine.Engine:
        """Get one SQLAlchemy Engine by connection str or metakey"""
        pass

    @abstractmethod
    def query(self, connection: str = None) -> DbQueryBuilder[DbQueryBuilder, Any]:
        """Database query builder passthrough"""


    @abstractmethod
    async def execute(
        self,
        query: Any,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.CursorResult:
        """Execute a SQLAlchemy Core Query based on connection str or metakey"""
        pass


    @abstractmethod
    async def all(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Sequence[sa.Row]:
        """Get many records from query. Returns empty List if no records found"""
        pass

    @abstractmethod
    async def fetchall(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Sequence[sa.Row]:
        """Alias to .all()"""
        pass

    @abstractmethod
    async def first(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row|None:
        """Get one (first/top) record from query. Returns None if no records found"""
        pass

    @abstractmethod
    async def fetchone(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row|None:
        """Alias to .first()"""
        pass

    @abstractmethod
    async def one(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row:
        """Get one record from query. Throws Exception if no data found or querying more than one record"""
        pass

    @abstractmethod
    async def one_or_none(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row|None:
        """Get one record from query.  Returns None if no record found.  Throws Exception of querying more than one record"""
        pass

    @abstractmethod
    async def scalars(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Sequence[Any]:
        """Get many scalar values from query.  Returns empty List if no records found. If selecting multiple columns, returns List of FIRST column only."""
        pass

    @abstractmethod
    async def scalar(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Any|None:
        """Get a single scalar value from query. Returns None if no record found.  Returns first (top) if more than one record found"""
        pass

    @abstractmethod
    async def scalar_one(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Any:
        """Get a single scalar value from query.  Throws Exception if no data found or if querying more than one record"""
        pass

    @abstractmethod
    async def scalar_one_or_none(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Any|None:
        """Get a single scalar value from query.  Returns None if no record found.  Throws Exception if querying more than one record"""
        pass

    @abstractmethod
    async def insertmany(self,
        query: sa.Insert|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> List[sa.Row]:
        """Bulk insert many rows, returning bulk primary keys (for databases that support INSERT..RETURNING)"""
        pass

    @abstractmethod
    async def insertone(self,
        query: sa.Insert|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row:
        """Insert one row, returning the one rows PK (as a tuple in case of dual PKs)"""
        pass



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import List, Optional, Any

class Relation(ABC):

    model: str
    foreign_key: Optional[str]
    local_key: Optional[str]
    name: Optional[str]
    entity: Optional[Any]

    @abstractmethod
    def fill(self, field: Field) -> Relation:
        pass

    @abstractmethod
    def is_one(self) -> bool:
        pass

    @abstractmethod
    def is_many(self) -> bool:
        pass

    @abstractmethod
    def is_type(self, *args) -> bool:
        pass

    @abstractmethod
    def contains_many(self, relations, skip: List = []) -> bool:
        pass


# At bottom due to circular issues between these two contracts
from .field import Field  # isort:skip



import prettyprinter
from prettyprinter import *

# Enable extras like @dataclass
prettyprinter.install_extras(exclude=[
     # Not using django, no need to print django models
    'django',

    'numpy',

    # Maybe when I get a shell ?
    'ipython_repr_pretty',
    'ipython',
    'attrs',
])



import uvicore
from uvicore.package import Provider
from uvicore.typing import Any, Dict
from uvicore.support.module import load
from uvicore.support.dumper import dump, dd
from uvicore.database.package import bootstrap
from uvicore.console.package.registers import Cli
from uvicore.foundation.events import app as AppEvents


@uvicore.provider()
class Database(Provider, Cli):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # Register IoC bindings
        # self.bind('Database', 'uvicore.database.db._Db',
        #     singleton=True,
        #     aliases=['database', 'db'],
        # )

        # Set uvicore.log global
        uvicore.db = uvicore.ioc.make('uvicore.database.db.Db')

        # Register event listeners
        AppEvents.Booted.listen(bootstrap.Database)

        # Event Handlers
        # String based events instead of class based because HTTP may not even
        # be installed, so importing the HTTP event would cause an issue.

        # Connect to all databases one time, after the system has started up
        # We also connect on the fly in db.py def database()

        # FIXME, NO databases is obsolete, this does nothing
        @uvicore.events.handle(['uvicore.console.events.command.Startup', 'uvicore.console.events.command.PytestStartup', 'uvicore.http.events.server.Startup'])
        async def uvicore_startup(event):
            # Connect to all databases
            pass
            #await uvicore.db.connect(all_dbs=True)

        # Disconnect from all databases after the system has shutdown
        @uvicore.events.handle(['uvicore.console.events.command.Shutdown', 'uvicore.console.events.command.PytestShutdown', 'uvicore.http.events.server.Shutdown'])
        async def uvicore_shutdown(event):
            # Disconnect from all connected databases
            pass
            #await uvicore.db.disconnect(all_dbs=True)

    def boot(self) -> None:
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define db commands
        self.register_cli_commands(
            group='db',
            help='Database Commands',
            commands={
                'create': 'uvicore.database.commands.db.create',
                'drop': 'uvicore.database.commands.db.drop',
                'recreate': 'uvicore.database.commands.db.recreate',
                'seed': 'uvicore.database.commands.db.seed',
                'reseed': 'uvicore.database.commands.db.reseed',
                'connections': 'uvicore.database.commands.db.connections',
            }
        )

        # Extend schematic generator commands
        self.register_cli_commands(
            group='gen',
            commands={
                'table': 'uvicore.database.commands.generators.table',
                'seeder': 'uvicore.database.commands.generators.seeder',
            }
        )



import uvicore
from uvicore.support.dumper import dd, dump
from uvicore.contracts import Config as ConfigInterface


@uvicore.service('uvicore.configuration.configuration.Configuration',
    aliases=['Configuration', 'Config', 'configuration', 'config'],
    singleton=True,
)
class Configuration(ConfigInterface):
    pass



# IoC Class Instance
# Not to be imported by the public from here.
# Use the uvicore.config singleton global instead.

# Public API for import * and doc gens
#__all__ = ['_Configuration']



import jwt
import uvicore
from uvicore.auth.user_info import UserInfo
from uvicore.support.hash import sha1
from uvicore.contracts import UserProvider
from uvicore.support.dumper import dump, dd
from uvicore.http.request import HTTPConnection
from uvicore.auth.support import password as pwd
from uvicore.typing import List, Union, Any, Dict, Optional, Callable
from uvicore.support import module


@uvicore.service()
class Jwt(UserProvider):
    """Retrieve user from JWT header during Authentication middleware

    This is a stateless, databaseless authenticator using contents from the JWT only!
    """

    async def _retrieve_user(self,
        key_name: str,
        key_value: Any,
        request: HTTPConnection,
        *,

        # Parameters from auth config
        anonymous: bool = False,
        anonymous_user: Dict = {},
        jwt_mapping: Dict = {},

        # Parameters from authenticator method
        jwt: Dict = {},
        role_permission_map: Optional[Dict] = None,

        # Must have kwargs for infinite allowed optional params, even if not used.
        **kwargs,
    ) -> UserInfo:
        """Retrieve user from backend"""

        if anonymous:
            # User is not logged in.
            # New anonymous user based on anonymous user configuration
            user = UserInfo(**anonymous_user)
            user.authenticated = False

        else:
            # New user based on JWT and mappings configuration
            user = UserInfo(
                id=jwt_mapping.id(jwt) or jwt.sub or 'Unknown',
                uuid=jwt_mapping.uuid(jwt) or jwt.sub or 'Unknown',
                sub=jwt_mapping.uuid(jwt) or jwt.sub or 'Unknown',
                username=jwt_mapping.username(jwt) or jwt.email or 'Unknown',
                email=jwt_mapping.email(jwt) or jwt.email or 'Unknown',
                first_name=jwt_mapping.first_name(jwt) or jwt.first_name or '',
                last_name=jwt_mapping.last_name(jwt) or jwt.last_name or '',
                title=jwt_mapping.title(jwt) or jwt.title or '',
                avatar=jwt_mapping.avatar(jwt) or jwt.avatar or '',
                groups=jwt_mapping.groups(jwt) or jwt.groups or [],
                roles=jwt_mapping.roles(jwt) or jwt.roles or [],
                permissions=jwt_mapping.permissions(jwt) or jwt.permissions or [],
                superadmin=jwt_mapping.superadmin(jwt) or jwt.admin or False,
                authenticated=True,
            )

        # If a role_permission_map Dict is passed, map user 'roles' into user
        # permissions for stateless scopes
        if role_permission_map is not None:
            # Import mapper passing in user object
            #mapper = module.load(role_permission_mapper).object()

            # Call __call__ on mapper
            #user.permissions = await mapper(user)

            permissions = []
            for role in user.roles:
                if role in role_permission_map:
                    permissions.extend(role_permission_map[role])
            user.permissions = permissions

        # Return user
        return user



import uvicore
import json as JSON
from uvicore.typing import Dict
from uvicore.console import command, argument, option
from uvicore import app, log
from uvicore.support.dumper import dd, dump

# Commands
# list = typer.Typer()
# show = typer.Typer()
# providers = typer.Typer()


@command()
def list():
    """List all packages"""
    log.header("List of all Packages (in exact order of registration dependency)")
    log.line()
    dump(app.packages)
    # for package in app.packages.values():
    #     dump("Package: " + package.name)
    #     dump(package)
    #     print()
        #dump(f"== {package.name} deep merged configs --")
        #dump(package.config())


@command()
@argument('package')
def get(package: str):
    """Show detailed info for one package"""
    if package == 'main':
        pkg = app.package(main=True)
    else:
        pkg = app.package(package)

    if pkg:
        log.header("Package object for " + package)
        log.line()
        dump(pkg)
        print()

        #log.header("Deep merged configs for " + package)
        #log.line()
        #dump(pkg.config())
    else:
        exit(f"Package {package} not found")

@command()
@option('--json', is_flag=True, help='Show providers as JSON')
def providers(json: bool):
    """Show providers graph"""
    if json:
        # This json stuff is just experimantal junk
        print(JSON.dumps(app.providers))
    else:
        log.header("Package provider graph (in exact order of registration dependency)")
        log.line()
        dump(app.providers)

        log.nl()
        log.line()
        log.header("Package providers as seen from the Ioc")
        bindings = {key:binding for (key, binding) in uvicore.ioc.bindings.items() if binding.type.lower() == 'provider'}
        dump(bindings)





import uvicore
import inspect
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Mapper as MapperInterface
from uvicore.support.collection import haskey, getvalue
from sqlalchemy.engine.result import Row as RowProxy


@uvicore.service()
class Mapper(MapperInterface):
    """Entity mapper for model->table or table->model conversions"""

    def __init__(self, entity, *args):
        # Mapper can be used from the model as a @classmethod
        # or as an instance (post.mapper)
        if inspect.isclass(entity):
            # Mapper is being accessed from a static class (Post.mapper())
            # Both entity and instance are the entity itself, there is no instance
            self.entity = entity
            self.instance = entity
        else:
            # Mapper is being accessed from a class instance
            # So the actual entity is the instance.__class__ (the metaclass)
            # while the instance is the model class itself
            self.entity = entity.__class__
            self.instance = entity
        self.args = args

    def column(self) -> str:
        """Convert a model field name into a table column name"""
        fieldname = self.args[0]
        field = self.entity.modelfields.get(fieldname)
        if field:
            return field.column
        return fieldname

    def field(self):
        """Convert a table column name into a model field name"""
        column = self.args[0]
        for field in self.entity.modelfields.values():
            if field.column == column:
                return field.name
        return column

    def model(self, perform_mapping: bool = True):
        """Convert a dict or List[dict] into a model or List[Model]

        Only maps table->model fields if perform_mapping = True, else assume already model fields.
        Works on a single record as dict - entity.mapper(DictModel).model()
        Works on a list of dict - entity.mapper(ListOfDictModel).model()
        Passes through if already a Model or List[Model]
        If mixed List of Dict and Model, converts all to Models
        """

        if self.args:
            # Model(s) are being passed in (entity.mapper(models).table())
            values = self.args[0]
        else:
            # No passed values, use self instance (model.mapper().table())
            values = self.instance

        single = False
        if type(values) != list:
            values = [values]
            single = True

        models = []
        for value in values:
            if type(value) == RowProxy:
                # Convert SQLAlchemy row to model
                models.append(self._row_to_model(value))

            elif type(value) == dict:
                # Convert dict to actual Model instance
                if perform_mapping:
                    # Values table columns need mapped to model fields
                    models.append(self._row_to_model(value))
                else:
                    # Assume values are already in model fields
                    models.append(self.entity(**value))
            else:
                # Already a model instance
                models.append(value)

        if single: return models[0]
        return models

    def table(self):
        """Convert an model instance into a dictionary matching the tables columns (column mapper enabled)

        Works on a single model - model.mapper().table() or entity.mapper(model).table()
        Or on a list of model instances - entity.mapper(ListOfModelInstances).tabel()
        Or on a list of model dictionaries - entity.mapper(ListOfModelDict).table()
        Does not convert or recurse into relations
        """
        if self.args:
            # Model(s) are being passed in (entity.mapper(models).table())
            models = self.args[0]
        else:
            # No passed models, use self instance (model.mapper().table())
            models = self.instance

        single = False
        if type(models) != list:
            models = [models]
            single = True

        tables = []
        for model in models:
            row = {}
            if type(model) == dict:
                # Model is a dict, convert to model instance
                model = self.entity(**model)

            for (field, value) in model.__dict__.items():
                field = self.entity.modelfields.get(field)
                if field.column and not field.read_only:
                    row[field.column] = value
            tables.append(row)

        # Return
        if single: return tables[0]
        return tables

                # bulk = []
        # for model in models:
        #     if type(model) == dict:
        #         # Value is a dictionary, not an actual model, convert to model
        #         model = entity(**model)
        #     #bulk.append(model.to_table())
        #     #bulk.append(model.mapper(model).table())
        #     bulk.append(model.mapper().table())

        # columns = {}
        # for (field, value) in self.instance.__dict__.items():
        #     field = self.entity.modelfields.get(field)
        #     if field.column and not field.read_only:
        #         columns[field.column] = value
        # return columns

    def _row_to_model(self, row = None):
        """Convert a single table row (SQLAlchemy RowProxy) or DICT of table into a model instance"""
        if not row: row = self.args[0]
        prefix = None
        if len(self.args) == 2: prefix = self.args[1]
        fields = {}

        for field in self.instance.__modelfields__.values():
            if not field.column and not field.evaluate: continue

            # NO, because we added an override of show_writeonly() on query builder
            # So this is handled in metaclass.py selectable_columns instead
            # NO if field.write_only: continue

            if field.evaluate:
                if type(field.evaluate) == dict:
                    # Evaluate is a Dict with callback and named parameters
                    eval_method = field.evaluate['method']
                    del field.evaluate['method']
                    #dump('tuple here', field.evaluate)
                    fields[field.name] = eval_method(row, **field.evaluate)

                elif type(field.evaluate) == tuple:
                    # Evaluate is a Tuple with callback and parameters
                    eval_method = field.evaluate[0]
                    fields[field.name] = eval_method(row, *field.evaluate[1:])

                else:
                    # Evaluate is a callback
                    fields[field.name] = field.evaluate(row)

            else:
                column = field.column
                if prefix: column = prefix + '__' + column
                # Use haskey instead of hasattr because data could be an SQLAlchemy row (which is class notation)
                # or a dictionary.  haskey and getvalue work the same on class models or dictionaries!
                if haskey(row, column):
                    fields[field.name] = getvalue(row, column)
        return self.entity(**fields)



    # # Field->Column
    # Mapper('slug').column() !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # # Column->Field
    # Mapper('unique_slug').field() !!!!!!!!!!!!!!!!!!!!!!!

    # # Model->TableDict
    # Mapper(User(email='adsf')).table() !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # # List[Model]->List[TableDict]
    # Mapper([
    #     User(email='asdf'),
    #     User(email='adsfasdf'),
    # ]).table()

    # # TableDict->Model
    # Mapper({
    #     'email': 'asdf'
    # }).model()

    # # List[TableDict]->List[Model]
    # Mapper([
    #     {
    #         'email': 'adsf',
    #     },
    #     {
    #         'email': 'asdasdf',
    #     }
    # ]).model()

    # # Table name - no, should be in db, has nothing todo with entities
    # Mapper('auth.users').tablename()



    # Usage:
    # Post.mapper().table()
    # post.mapper().table()



from pydantic import BaseModel
from uvicore.support.printer import pretty_call, register_pretty


class JobResults(BaseModel):
    # There are no default properties here
    # So why a JobResults at all?  Basically to
    # wrap pydantic BaseModel and provide better pretty priting
    pass


@register_pretty(JobResults)
def pretty_entity(value, ctx):
    """Custom pretty printer for JobResults"""
    return pretty_call(ctx, value.__class__, **value.__dict__)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, Any

class Cache(ABC):

    @abstractmethod
    def connect(self, store: str = None) -> Cache:
        """Connect to a cache backend store"""

    @abstractmethod
    def store(self, store: str = None) -> Cache:
        """Alias to connect"""

    @abstractmethod
    async def has(self, key: str) -> bool:
        """Check if key exists"""

    @abstractmethod
    async def get(self, key: Union[str, List], *, default: Any = None) -> Any:
        """Get one or more key values if exists else return default value"""

    @abstractmethod
    async def remember(self, key: str, callback: Callable, *, seconds: int = 0) -> Any:
        """Get a key if exists, if not SET the key to callback value"""

    @abstractmethod
    async def put(self, key: Union[str, Dict], value: Any = None, *, seconds: int = 0) -> None:
        """Put one or more key/values in cache with optional expire in seconds (0=never expire)"""

    @abstractmethod
    async def pull(self, key: str) -> Any:
        """Get a value from cache them remove it"""

    @abstractmethod
    async def add(self, key: str, value: Any, *, seconds: int = 0) -> bool:
        """Put a single value in cache only if not exists"""

    @abstractmethod
    async def touch(self, key: str, *, seconds: int = None) -> bool:
        """Touch a key, if seconds are provided, also reset expire TTL"""

    @abstractmethod
    async def increment(self, key, by: int = 1, *, seconds: int = None) -> int:
        """Increment a key by integer specified.  If key not exists, sets key to increment"""

    @abstractmethod
    async def decrement(self, key, by: int = 1, *, seconds: int = None) -> int:
        """Decrement a key by integer specified.  If key not exists, sets key to decrement value"""

    @abstractmethod
    async def forget(self, key: str) -> None:
        """Delete a key from cache"""

    @abstractmethod
    async def flush(self) -> None:
        """Flush entire cache.  Only deletes keys with proper cache prefix."""



from __future__ import annotations
from abc import ABC, abstractmethod
from uvicore.typing import List, Dict, Callable, TypeVar, Generic, Decorator, Optional, Any, Union

# Generic Route (Web or Api)
R = TypeVar('R')

# This is my new custom uvicore router

class WebRoute(Dict):
    """WebRoute superdict definition"""

    # These class level properties for for type annotations only.
    # They do not restrict of define valid properties like a dataclass would.
    # This is still a fully dynamic SuperDict!
    path: str
    name: str
    endpoint: Callable
    methods: List[str]
    middleware: Optional[List]
    original_path: str
    original_name: str


class ApiRoute(Dict):
    """WebRoute superdict definition"""

    # These class level properties for for type annotations only.
    # They do not restrict of define valid properties like a dataclass would.
    # This is still a fully dynamic SuperDict!
    path: str
    name: str
    prefix: bool
    endpoint: Callable
    methods: List[str]
    response_model: Optional[Any]
    response_class: Optional[Any]
    tags: Optional[List]
    middleware: Optional[List]
    summary: Optional[str]
    description: Optional[str]
    original_path: str
    original_name: str


class Routes(ABC):
    """Routes and Controller Class"""

    middleware = None
    auth = None
    scopes = None

    @property
    @abstractmethod
    def package(self) -> Package: pass


class Router(Generic[R], ABC):
    """Abstract base router class for Web and Api Router Implimentations"""

    @property
    @abstractmethod
    def package(self) -> Package: pass

    @property
    @abstractmethod
    def routes(self) -> Dict[str, R]: pass

    @abstractmethod
    def controller(self,
        module: Union[str, Callable],
        *,
        prefix: str = '',
        name: str = '',
        tags: Optional[List[str]] = None,
        options: Dict = {}
    ) -> List:
        """Include a Route Controller"""

    @abstractmethod
    def include(self,
        module: Union[str, Callable],
        *,
        prefix: str = '',
        name: str = '',
        tags: Optional[List[str]] = None,
        options: Dict = {}
    ) -> List:
        """Alias to controller"""

    @abstractmethod
    def group(self, prefix: str = '', *,
        routes: Optional[List] = None,
        name: str = '',
        tags: Optional[List[str]] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Any] = None,
        scopes: Optional[List] = None,
    ) -> Callable[[Decorator], Decorator]:
        """Route groups method and decorator"""

class WebRouter(ABC):
    # Fixme, add get, post, put...
    pass


class ApiRouter(ABC):
    # Fixme, add get, post, put...
    pass


class ModelRouter(ABC):
    # Fixme, add get, post, put...
    pass


# Bottom to avoid circular dependencies
from .package import Package  # isort:skip

















# These were old, for fastapi and starlettee direct

# class WebRouter(ABC):
#     """Router for Web Controllers"""

#     @property
#     @abstractmethod
#     def router(self) -> Any: pass

#     @property
#     @abstractmethod
#     def routes(self) -> List[BaseRoute]: pass

#     @property
#     @abstractmethod
#     def on_startup(self) -> None: pass

#     @property
#     @abstractmethod
#     def on_shutdown(self) -> None: pass

#     @abstractmethod
#     def get(self,
#         path: str,
#         name: str = None,
#     ) -> Callable: pass

#     @abstractmethod
#     def include_router(self, router: "WebRouter") -> None: pass


# class ApiRouter(ABC):
#     """Router for API Endpoints"""

#     @property
#     @abstractmethod
#     def router(self) -> Any: pass

#     @property
#     @abstractmethod
#     def routes(self) -> List[BaseRoute]: pass

#     @property
#     @abstractmethod
#     def on_startup(self) -> None: pass

#     @property
#     @abstractmethod
#     def on_shutdown(self) -> None: pass

#     @abstractmethod
#     def get(self,
#         path: str,
#         name: str = None,
#         *,
#         response_model: Type[Any] = None,
#         include_in_schema: bool = True
#     ) -> Callable: pass

#     @abstractmethod
#     def include_router(self, router: "APIRouter") -> None: pass



import os
import uvicore
import aiohttp
import aiofiles
from uvicore.typing import Dict, List
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Email
from aiohttp import BasicAuth


@uvicore.service()
class Mailgun:

    @classmethod
    async def send(cls, message: Email, options: Dict):
        # Get aiohttp ClientSession
        http = uvicore.ioc.make('http_client')

        # Get message body as text or html
        body_type = 'text'
        body_content = ''
        if message.html:
            body_type = 'html'
            body_content = message.html
        elif message.text:
            body_type = 'text'
            body_content = message.text

        # New multi-part form data (because mailgun can accept both email fields and attachment data)
        data = aiohttp.FormData()

        # Add from name and address
        data.add_field('from', message.from_name + '<' + message.from_address + '>')

        # Add to recipients
        for to in message.to:
            data.add_field('to', to)

        # Add cc recipients
        for cc in message.cc:
            data.add_field('cc', cc)

        # Add bcc recipients
        for bcc in message.bcc:
            data.add_field('bcc', bcc)

        # Add subject
        data.add_field('subject', message.subject)

        # Add body (as html or text)
        data.add_field(body_type, body_content)

        # Add attachments
        for attachment in message.attachments:
            if os.path.exists(attachment):
                filename = attachment.split('/')[-1]
                data.add_field('attachment', await aiofiles.open(attachment, 'rb'), filename=filename)

        # Post to mailgun using async aiohttp
        async with http.post(
            url='https://api.mailgun.net/v3/' + options.domain + '/messages',
            auth=BasicAuth('api', options.secret),
            data=data,
        ) as r:
            # Success
            detail = await r.text()
            if r.status == 200:
                return

            # Failure
            raise Exception("Could not send mailgun email - " + detail)



#from fastapi.routing import APIRouter
#from fastapi_utils.inferring_router import InferringRouter


from typing import TYPE_CHECKING, Any, Callable, get_type_hints
from fastapi import APIRouter
class InferringRouter(APIRouter):
    """
    Overrides the route decorator logic to use the annotated return type as the `response_model` if unspecified.
    """

    if not TYPE_CHECKING:  # pragma: no branch

        def add_api_route(self, path: str, endpoint: Callable[..., Any], **kwargs: Any) -> None:
            if kwargs.get("response_model") is None:
                kwargs["response_model"] = get_type_hints(endpoint).get("return")
            return super().add_api_route(path, endpoint, **kwargs)

class Router(InferringRouter):
    pass





import uvicore
from uvicore.typing import Callable, Decorator


def event(name: str = None) -> Callable[[Decorator], Decorator]:
    def decorator(cls: Decorator) -> Decorator:
        # Bind this Event into the Ioc
        bind_name = name or cls.__module__ + '.' + cls.__name__
        new_cls = uvicore.ioc.bind_from_decorator(cls, name=bind_name, object_type='event', singleton=False)

        # Other things with an event

        return new_cls
    return decorator



import uvicore
from uvicore.http import Request
from uvicore.http.routing import Routes, ApiRouter, ModelRouter, Guard
from uvicore.support.dumper import dump, dd
from uvicore.auth.user_info import UserInfo

@uvicore.routes()
class Api(Routes):

    # Apply scopes to all routes and children controllers
    #scopes = ['authenticated']
    #scopes = ['authenticated', 'employee']
    #scopes = None

    def register(self, route: ApiRouter):
        """Register API Route Endpoints"""

        # Define controller base path
        route.controllers = 'uvicore.auth.http.api'

        @route.group(tags=['Auth'], scopes=['authenticated'])
        def authenticated_routes():

            @route.get('/userinfo')
            def userinfo(request: Request) -> UserInfo:
                """Detailed User Info Including Roles and Permissions"""

                # HTTP example
                # TOKEN=$(fa-api tgb-local login wiki-vue-app token)
                # http GET 'https://wiki-api-local.triglobal.io/api/auth/userinfo' Authorization:"Bearer $TOKEN"

                # Get the user from the request
                user: UserInfo = request.scope['user']
                # uvicore.auth.user_info.UserInfo(
                #     id=2,
                #     uuid='823003ad-6e1f-42ed-a024-f45f400c1b30',
                #     username='mreschke@triglobalblockchain.com',
                #     email='mreschke@triglobalblockchain.com',
                #     first_name='Matthew',
                #     last_name='Reschke',
                #     title='',
                #     avatar='',
                #     groups=['Administrator'],
                #     roles=['Administrator'],
                #     permissions=['authenticated', 'admin'],
                #     superadmin=True,
                #     authenticated=True
                # )

                # Return UserInfo
                return user

        # Return router
        return route



from abc import ABC, abstractmethod
from typing import Callable, List, Any
try:
    from starlette.types import ASGIApp
except ImportError:
    pass


class Server(ABC):

    @property
    @abstractmethod
    def server(self) -> Any: pass

    @abstractmethod
    def include_router(self, router, *, prefix: str = '', tags: List[str] = None) -> None: pass

    @abstractmethod
    def mount(self, path: str, app: ASGIApp, name: str = None) -> None: pass

    @abstractmethod
    def on_event(self, event_type: str) -> Callable: pass




import uvicore
from typing import List
from starlette.staticfiles import StaticFiles as _Static


@uvicore.service('uvicore.http.static.StaticFiles', aliases=['StaticFiles'])
class StaticFiles(_Static):

    def __init__(self, directories: List[str]):
        super().__init__()

        # Starlette static files has a reverse winning directory order
        # In starlette the FIRST defined directory wins.  But in uvicore
        # the LAST defined provider always wins (for configs, views...)
        # So we have to reverse the list
        self.all_directories = [x for x in reversed(directories)]

    def get_directories(self, directories, packages):
        pass


# IoC Class Instance
# No because not to be used by the public
#StaticFiles: _StaticFiles = uvicore.ioc.make('StaticFiles')

# Public API for import * and doc gens
#__all__ = ['_StaticFiles']



import uvicore
from starlette.applications import Starlette as _Starlette
from uvicore.support.dumper import dd, dump


@uvicore.service('uvicore.http.servers.web._Server',
    #aliases=['api', 'web_server'],
    singleton=True,
    kwargs={
        'debug': uvicore.config('app.debug'),
    },
)
class Server(_Starlette):
    @property
    def server(self) -> _Starlette:
        return self



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from .connection import Connection
from uvicore.typing import Dict, List, Any
from uvicore.support.printer import pretty_call, register_pretty


class Package(Dict):
    """Package Definition"""

    # These class level properties are for type annotations only.
    # They do not restrict of define valid properties like a dataclass would.
    # This is still a fully dynamic SuperDict!
    name: str
    short_name: str
    vendor: str
    version: str
    main: bool
    path: str
    paths: Dict
    registers: Dict
    database: _Database
    web: _Web
    api: _Api
    console: _Console

    @abstractmethod
    def config(self, dotkey: str = None) -> Dict:
        """Helper to access this packages configs"""
        pass

    @abstractmethod
    def connection(self, name: str = None) -> Connection:
        """Helper to access this packages connections"""
        pass


class _Database(Dict):
    """Simple Package Database Type Annotations for the SuperDict"""
    connections: List[Connection]
    connection_default: str
    models: List[str]
    tables: List[str]
    seeders: List[str]


class _Web(Dict):
    prefix: str
    routes_module: str
    routes: Dict[str, WebRoute]
    view_paths: List[str]
    asset_paths: List[str]
    template_options: Dict

class _Api(Dict):
    prefix: str
    routes_files: str
    routes: Dict[str, ApiRoute]


class _Http(Dict):
    """Simple Package HTTP Type Annotations for the SuperDict"""
    web_routes: List[str]
    api_routes: List[str]
    web_route_prefix: str
    api_route_prefix: str
    middleware: List[str]
    view_paths: List[str]
    asset_paths: List[str]
    template_options: Dict



class _Console:
    """Simple Package Console Type Annotations for the SuperDict"""
    groups: Dict


# Bottom to avoid circular dependencies
from .router import WebRoute, ApiRoute  # isort:skip



# @dataclass
# class Package_OBSOLETE_NOW_DIC(ABC):
#     """asdfasdf"""
#     name: str
#     location: str
#     main: bool
#     #registers: Registers  # Register (singular) is a reserved word in ABC meta
#     web_route_prefix: str
#     api_route_prefix: str
#     view_paths: List[str]
#     asset_paths: List[str]
#     template_options: Dict[str, str]
#     connection_default: str
#     connections: List[Connection]
#     #models: List[str]
#     seeders: List[str]
#     commands: Dict
#     config: Dict

#     @abstractmethod
#     def config(self, dotkey: str = None) -> Any:
#         pass

#     @abstractmethod
#     def connection(self, name: str = None) -> Connection:
#         pass



@register_pretty(Package)
def pretty_entity(value, ctx):
    """Custom pretty printer for my SuperDict"""
    # This printer removes the class name uvicore.types.Dict and makes it print
    # with a regular {.  This really cleans up the output!

    # SuperDict are printed as Dict, but this Package SuperDict should
    # be printed more like a class with key=value notation, so use **values
    return pretty_call(ctx, 'Package', **value)



from abc import ABC, abstractmethod
from uvicore.typing import Union, List, Dict
from uvicore.contracts.user_info import UserInfo
try:
    from starlette.requests import HTTPConnection
except:
    HTTPConnection = None


class UserProvider(ABC):

    def __init__(self):
        # Default field mapping for each method below, easily overridable in your custom user providers!
        self.field_map = {
            'id': 'id',
            'uuid': 'uuid',
            'username': 'username',
        }

    async def retrieve_by_id(self, id: Union[str, int], request: HTTPConnection, **kwargs) -> UserInfo:
        """Retrieve user by primary key from the user provider backend.  No validation."""
        field = self.field_map['id']
        return await self._retrieve_user(field, id, request, **kwargs)

    async def retrieve_by_uuid(self, uuid: str, request: HTTPConnection, **kwargs) -> UserInfo:
        """Retrieve the user by uuid from the user provider backend.  No validation."""
        field = self.field_map['uuid']
        return await self._retrieve_user(field, uuid, request, **kwargs)

    async def retrieve_by_username(self, username: str, request: HTTPConnection, **kwargs) -> UserInfo:
        """Retrieve the user by username from the user provider backend.  No validation."""
        field = self.field_map['username']
        return await self._retrieve_user(field, username, request, **kwargs)

    async def retrieve_by_credentials(self, username: str, password: str, request: HTTPConnection, **kwargs) -> UserInfo:
        """Retrieve the user by username from the user provider backend AND validate the password if not None"""
        field = self.field_map['username']
        return await self._retrieve_user(field, username, request, password=password, **kwargs)

    @abstractmethod
    async def create_user(self, request: HTTPConnection, **kwargs):
        """Create new user in backend"""

    @abstractmethod
    async def sync_user(self, request: HTTPConnection, **kwargs):
        """Sync user to backend"""

    # @abstractmethod
    # async def retrieve_by_id(self, id: Union[str, int], request: HTTPConnection, options: Dict = {}) -> UserInfo:
    #     """Retrieve user by primary key from the user provider backend.  No validation."""

    # @abstractmethod
    # async def retrieve_by_uuid(self, uuid: str, request: HTTPConnection, options: Dict = {}) -> UserInfo:
    #     """Retrieve the user by uuid from the user provider backend.  No validation."""

    # @abstractmethod
    # async def retrieve_by_username(self, username: str, request: HTTPConnection, options: Dict = {}) -> UserInfo:
    #     """Retrieve the user by username from the user provider backend.  No validation."""

    # @abstractmethod
    # async def retrieve_by_credentials(self, username: str, password: str, request: HTTPConnection, options: Dict = {}) -> UserInfo:
    #     """Retrieve the user by username from the user provider backend AND validate the password if not None"""



import uvicore
from uvicore.typing import Dict
from uvicore.events import Handler
from uvicore.support.dumper import dump, dd
from uvicore.foundation.events.app import Booted as OnAppBooted


class Redis(Handler):

    def __call__(self, event: OnAppBooted):
        """Bootstrap Redis after the Application is Booted"""

        connections = Dict()
        last_default = None; app_default = None
        for package in uvicore.app.packages.values():
            if not 'redis' in package: continue

            # Get last defined default connection
            if package.redis.connection_default: last_default = package.redis.connection_default

            # Get running app default connection
            if package.main and package.redis.connection_default: app_default = package.redis.connection_default

            # Append connections
            connections.merge(package.redis.connections)

        # Add all final merged connections to running_config
        uvicore.app.add_running_config('redis.connections', connections)

        # Initialize Redis with all connections at once
        redis = uvicore.ioc.make('uvicore.redis.redis.Redis')
        redis.init(app_default or last_default, connections)



# type: ignore
from .provider import Provider
#from uvicore.package.package import Package as _Package

# Proxy instead of import to create a nicer uvicore.package.Package namespace
# Even if you override this the acutal class, all listings will still show this consistent namespace
# NO - Not sure I like proxy as it creates 2 classes on 2 different paths which can cause issues in type()
# class Package(_Package):
#     pass

from .package import Package



import uvicore
from uvicore.typing import Dict
from uvicore.http import Request
from uvicore.http.exceptions import HTTPException
from uvicore.http.routing import ApiRouter, Controller

# Extra
# from uvicore.auth import UserInfo
# from uvicore.http.routing import Guard
# from uvicore.typing import Dict, List, Optional
# from uvicore.http.exceptions import HTTPException
# from uvicore.http.params import Path, Query, Header, Cookie, Body, Form, File, Depends, Security


# ------------------------------------------------------------------------------
# Uvicore Controller Schematic
# This schematic is filled with examples, a suppliment to the docs.
# Pick the best example for your use case and modify as needed!
# ------------------------------------------------------------------------------


@uvicore.controller()
class xx_ControllerName(Controller):

    # --------------------------------------------------------------------------
    # Example:  Multiple ways to apply route level middleware including
    # auth guards with scoped permissions to this entire controller.
    # Tip: Your routes/api.py and routes/web.py are actually the same
    # as controllers.  In fact controllers are just nested routers.  So all
    # notes here also apply to your routes/* files and controllers.
    # --------------------------------------------------------------------------
    # Apply scopes to all routes and children controllers - simple and preferred
    #scopes = ['authenticated', 'employee']

    # Scopes is a shortcut to
    #auth = Guard(['authenticated', 'gardner'])

    # Scopes and auth are shortcuts to the full middleware stack
    # middleware = [
    #     Guard(['authenticated', 'post_manager']),
    #     Any other route based middleware here
    # ]

    # Optionally, apply scopes and also grab the current user
    # user: UserInfo = Guard(['authenticated'])
    #   Then in your routes, inject the user with
    #   async def welcome(request: Request, user: UserInfo = self.user):

    def register(self, route: ApiRouter):
        """Register API Controller Endpoints"""

        # ----------------------------------------------------------------------
        # Example: Basic route responding with a view dict to JSON blob
        # ----------------------------------------------------------------------
        @route.get('/example1', tags=['Examples'])
        async def example1() -> Dict:
            """This docstring shows up in openapi"""
            try:
                return {'welcome': 'to uvicore API!'}
            except Exception as e:
                raise HTTPException(500, exception=e)


        # ----------------------------------------------------------------------
        # Example: Route returning a Model schema with python return type hint
        # ----------------------------------------------------------------------
        # @route.get('/example2', tags=['Examples'])
        # async def example2()) -> List[models.Post]:
        #     """This docstring shows up in openapi"""
        #     return await models.Post.query().get()


        # ----------------------------------------------------------------------
        # Example: Route returning a Model schema using response_model
        # ----------------------------------------------------------------------
        # @route.get('/example3', response_model=List[models.Post], tags=['Examples'])
        # async def example3():
        #     return await models.Post.query().get()


        # ----------------------------------------------------------------------
        # Example: Auth guard using scopes shortcut
        # ----------------------------------------------------------------------
        # @route.get('/example4/{id}', scopes=['authenticated'], tags=['Examples'])
        # async def example4(id: int) -> models.Post:
        #     return await models.Post.query().find(id)


        # ----------------------------------------------------------------------
        # Example: Auth guard using auth shortcut
        # ----------------------------------------------------------------------
        # @route.get('/example5/{id}', auth=Guard(['authenticated']), tags=['Examples'])
        # async def example5(id: int) -> models.Post:
        #     return await models.Post.query().find(id)


        # ----------------------------------------------------------------------
        # Example: Auth guard while also getting the current user
        # Also accepts an optional GET parameter (?name=matthew)
        # ----------------------------------------------------------------------
        # @route.get('/example6/{id}', tags=['Examples'])
        # async def example6(id: int, name: Optional[str], user: UserInfo = Guard(['authenticated'])) -> models.Post:
        #     dump(name, user)
        #     return await models.Post.query().find(id)


        # ----------------------------------------------------------------------
        # Example: Auth guard using middleware
        # Get the current user with request.scope['user']
        # Also accepts an optional GET parameter (?name=matthew)
        # ----------------------------------------------------------------------
        # @route.get('/example6a/{id}', middleware=[
        #     Guard(['authenticated', 'manager']),
        #     # Any other route based middleware here
        # ], tags=['Examples'])
        # async def example6a(request: Request, id: int, name: Optional[str]) -> models.Post:
        #     user = request.scope['user']
        #     dump(name, user)
        #     return await models.Post.query().find(id)

        # ----------------------------------------------------------------------
        # Example: Other types of responses
        # ----------------------------------------------------------------------
        # @route.get('/example6b')
        # async def example6b(request: Request):
        #     return response.Text('Text Here')
        #     return response.HTML('<b>HTML</b> here')
        #     return response.JSON({'json':'here'})
        #     return response.UJSON({'json':'here'}) # requires ujson dependency
        #     # and more ... see uvicore/http/response.py

        # ----------------------------------------------------------------------
        # Example: Changing the route name
        # Each route is given a name automatically.  A name is what you can use
        # to reference the route in your code and view templates.  Try to use the
        # route name instead of the path as paths WILL change as other users use
        # your library because they can tweak the BASE PATH.  Views should be using
        # {{ url('xx_appname.ex7')}}, never /example7
        # If you don't specify a name, uvicore makes a name automatically
        # from the path, even nested paths from groups.  Name always starts
        # with your apps name, ie: xx_appname.
        # ----------------------------------------------------------------------
        # @route.get('/example7', name='ex7')
        # async def example2a(request: Request):
        #     # Route name is xx_appname.ex7 instead of default xx_appname.example7
        #     return response.Text('example7')


        # ----------------------------------------------------------------------
        # Example: Override the naming autoprefix
        # When you define a custom route name, a prefix of xx_appname. is
        # automatically added keeping your route scoped to this package.
        # In order to set a full name you must set autoprefix=False.
        # This is handy when you want to override a route from another package.
        # ----------------------------------------------------------------------
        # @route.get('/example8', name='someother.app.ex8', autoprefix=False)
        # async def example8(request: Request):
        #     # Route name is someother.app.ex8
        #     return response.Text('example8')


        # ----------------------------------------------------------------------
        # Example: POST a model with validation
        # ----------------------------------------------------------------------
        # @route.post('/example9')
        # async def example9(post: models.Post):
        #     models.Post.insert(post)


        # ----------------------------------------------------------------------
        # Example raise proper HTTP Exception
        # ----------------------------------------------------------------------
        # @route.get('/example9a')
        # async def example9a():
        #     raise HTTPException(404, 'bad stuff')


        # ----------------------------------------------------------------------
        # Example: Grouping routes for common paths and scopes
        # Routes will be /group1/example10 and /group1/subgroup1/example11
        # with scopped permissions on both routes.
        # IF you also set name='g1', all route names will be
        # xx_appname.g1.example19 instead of autonamed xx_appname.group1.example10
        # ----------------------------------------------------------------------
        # @route.group('/group1', scopes=['authenticated'], tags=['Group'])
        # def group1():
        #     # Route will be under both Group and Example TAG
        #     @route.get('/example10', tags=['Example'])
        #     async def example10(request: Request):
        #         return response.Text("example10")

        #     @route.group('/subgroup1')
        #     def subgroup1():
        #         @route.get('/example11')
        #         async def example11(request: Request):
        #             return response.Text("example11")


        # ----------------------------------------------------------------------
        # Example: Routes as method callbacks (no decorators)
        # ----------------------------------------------------------------------
        # def example12(request: Request):
        #     return response.Text('example12')
        # route.get('/example12', example12)


        # ----------------------------------------------------------------------
        # Example: Groups and routes as method callbacks (no decorators)
        # ----------------------------------------------------------------------
        # def example13(request: Request, id: int) -> models.Post:
        #     return await models.Post.query().find(id)
        # def example14(request: Request, email: str) -> List[models.Post]:
        #     return await models.Post.query().where('email', email).get()
        # route.group('/group2', scopes=['authenticated', 'post_manager'], routes=[
        #     route.get('/example13', example13),
        #     route.get('/example14', example14),
        # ])


        # ----------------------------------------------------------------------
        # Example: Including other route files and controllers
        # Technically, there is no difference between your routes files
        # in http/routes/web.py and http/routes/api.py and your controllers
        # or api controllers.  They are all just one huge nested router.
        # They are only split up for logical convenience.  Just as you included
        # controlers from your http/routes/web.py, you can also include other
        # routes here.
        # ----------------------------------------------------------------------
        # route.controller('xx_vendor.xx_appname.http.controllers.some.other.Other')

        # Also, route.controller and route.include are aliases of each other, same thing.
        # route.include('xx_vendor.xx_appname.http.controllers.some.other2.Other2')

        # Instead of typing the full module path, if route.controllers is defined
        # Then all .controller() and .include() can use relative paths
        # route.controllers = 'xx_vendor.xx_appname.http.controllers'

        # Looks for Class in xx_vendor.xx_appname.http.controllers.some.Some
        # route.controller('some')

        # Leading period means APPEND path to defined route.controllers
        # So this looks for xx_vendor.xx_appname.http.controllers.other3.Other3
        # route.controller('.other3.Other3')

        # If no leading . but other . exists, then it is assuming a full path,
        # regardless if route.controllers is defined or not.


        # Return router
        # Must always return the router at the end of every controller and routes file
        # as this is one infinitely recursive nested router configuration.
        return route



import uvicore
from uvicore.console import click, group
from uvicore.console.events import command as ConsoleEvents

title_ideas = """
    The Full Stack Asynchronous Python Framework with the performance of FastAPI and the elegance of Laravel!
"""

# Must include actual bind name when stacking decorators as it cannot deduce on its own.
# Bind decorator must come first in the stack
@uvicore.service('uvicore.console.console.cli', aliases=['Console', 'console', 'cli'])
@group(help=f"""
    \b
    Uvicore {uvicore.__version__}
    The Fullstack Async Web, API and CLI Python Framework

    Copyright (c) 2024 Matthew Reschke
    License http://mreschke.com/license/mit
""")
@click.version_option(version=uvicore.__version__, prog_name='Uvicore Framework', flag_value='--d')
@click.pass_context
async def cli(ctx):
    # Console startup handler
    # This runs before any command.  Even if you don't actually run a command but only display
    # the help output using NO parameters, like ./uvicore http  It does not however run when just
    # showing ./uvicore main help output.  Only on subcommand help output or actual command
    await console_startup()

    # Console shutdown handler
    ctx.call_on_close(console_shutdown)

    # The actual command runs right here


# Had to hack asyncclick core.py def close and __aexit__ with awaits
# github issue open, hopefully he will fix soon


async def console_startup():
    #print('console startup')
    await ConsoleEvents.Startup().codispatch()


#@cli.resultcallback()
#async def console_shutdown(result, **kwargs):
#This decorator works only if you give it an actual command
#Just showing help output from NO command does not fire this, so aiohttp still complains
#about not being shutdown properly even if you just do ./uvicore http with no actual command to run.

async def console_shutdown():
    #print('console shutdown')
    await ConsoleEvents.Shutdown().codispatch()



class Singleton(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]



from uvicore.http import Request
from uvicore.http import response
from uvicore.http.exceptions import HTTPException


async def api(request: Request, e: HTTPException) -> response.JSON:
    """Main exception handler for all API endpoints"""

    # Get error payload (smart based on uvicore or stock HTTPException)
    (status_code, detail, message, exception, extra, headers) = expand_payload(e)
    return response.JSON(
        {
            "status_code": status_code,
            "message": message,
            "detail": detail,
            "exception": exception,
            "extra": extra,
        }, status_code=status_code, headers=headers
    )


async def web(request: Request, e: HTTPException) -> response.HTML:
    """Main exception handler for all Web endpoints"""

    # Get error payload (smart based on uvicore or stock HTTPException)
    (status_code, detail, message, exception, extra, headers) = expand_payload(e)

    try:
        # Try to respond with a errors template, if exists
        return await response.View('errors/' + str(status_code) + '.j2', {
            'request': request,
            **e.__dict__,
        })
    except:

        try:
            # Try to respond with a catch_all template, if exists
            return await response.View('errors/catch_all.j2', {
                'request': request,
                **e.__dict__,
            })
        except:
            # Errors status_code or catch_all template does not exist.
            # Response with generic HTML error
            html = f"""
            <div class="error">
                <h1>{status_code} {message}</h1>
                <p>{detail or ''}</p>

                <h3>Exception:</h3>
                <p>{exception or ''}</p>

                <h3>Extra:</h3>
                <pre>{extra or ''}</pre>
            </div>
            """
            return response.HTML(
                content=html,
                status_code=status_code,
                headers=headers
            )


def expand_payload(e: HTTPException):
    """Get error payload depending on uvicore or stock HTTPException"""
    # Defined in the running app config api.exceptions.main
    headers = getattr(e, "headers", None)

    # Base only has status_code and detail
    status_code = e.status_code
    detail = e.detail
    exception = None
    if (hasattr(e, 'message')):
        # Error called using uvicores new and expanded HTTPException
        message = e.message
        exception = e.exception
        extra = e.extra
    else:
        # Error called using base HTTPException
        message = e.detail
        extra = None
    return (status_code, detail, message, exception, extra, headers)



# Public API used in packages routes and controllers

# type: ignore
from .api_router import ApiRoute, ApiRouter
from .guard import Guard
from .router import Router
from .router import Routes
from .router import Routes as Controller
from .web_router import WebRoute, WebRouter

# Only import these if DB is installed
try:
    import sqlalchemy as sa
    from .auto_api import AutoApi
    from .model_router import ModelRouter
except ImportError:
    pass



from .application import Application
from .cache import Cache
from .config import Config
from .dispatcher import Dispatcher
from .ioc import Binding, Ioc
from .job_dispatcher import JobDispatcher
from .job import Job
from .logger import Logger
from .package import Package
from .provider import Provider
from .template import Template
from .user_info import UserInfo
from .user_provider import UserProvider
from .authenticator import Authenticator
from .email import Email

# Connection is used in package/package.py so not not optional
from .connection import Connection


# Optional imports based on installed modules
try:
    import sqlalchemy
    from .builder import DbQueryBuilder, OrmQueryBuilder, QueryBuilder
    from .database import Database
    from .field import Field
    from .mapper import Mapper
    from .model import Model
    from .relation import Relation
    from .auto_api import AutoApi
except ImportError:
    pass

try:
    import fastapi
    from .router import (ApiRoute, ApiRouter, ModelRouter, Router, Routes, WebRoute, WebRouter)
    from .server import Server
except ImportError:
    pass



#from .event import Event
#from .router import ApiRouter, WebRouter
#from .routes import Routes


# The contracts package uses a __init__.py because users will import these
# methods from their own apps and we want a nicer import which looks
# like - from uvicore.contracts import Model



from environs import Env


# The configuration package uses a __init__.py because users will import these
# methods from their own configs and we want a nicer import which looks
# like - from uvicore.configuration import env

# New environs Env instance
env = Env()

# No, not used publically, and is so early that uvicore.ioc is still None
#from .configuration import Configuration


# Public API for import * and doc gens
# __all__ = [
#     'Env', 'env',
# ]



import uvicore
import asyncio
import functools
from uvicore.typing import Callable, TypeVar, Any
from uvicore.support.dumper import dump, dd

try:
    import contextvars  # Python 3.7+ only.
except ImportError:  # pragma: no cover
    contextvars = None  # type: ignore

T = TypeVar("T")

async def run_in_threadpool(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    loop = asyncio.get_event_loop()
    if contextvars is not None:  # pragma: no cover
        # Ensure we run in the same context
        child = functools.partial(func, *args, **kwargs)
        context = contextvars.copy_context()
        func = context.run
        args = (child,)
    elif kwargs:  # pragma: no cover
        # loop.run_in_executor doesn't accept 'kwargs', so bind them in here
        func = functools.partial(func, **kwargs)
    return await loop.run_in_executor(None, func, *args)



# Does not work as I need it to
# from asgiref.sync import async_to_sync
# def asyncmethodOBSOLETE_TEST(func):
#     """Convert an async method into a sync method if running synchronously"""
#     def decorator(*args, **kwargs):
#         do_async = False

#         if do_async == False:
#         #if not uvicore.app.is_async:
#             # If NOT running async (so from CLI) convert all async calls to sync
#             value = async_to_sync(func)(*args, **kwargs)



#             # if async_to_sync.launch_map:
#             #     # Event loop already exists, call function asynchronously
#             #     dump(async_to_sync.__dict__)
#             #     value = func(*args, **kwargs)
#             # else:
#             #     # No event loop started, create a new one
#             #     value = async_to_sync(func)(*args, **kwargs)

#             # try:
#             #     #dump(async_to_sync.__dict__)
#             #     value = async_to_sync(func)(*args, **kwargs)
#             # except RuntimeError:
#             #     # Exception means event loop is already running
#             #     # So don't async_to_sync again or you get RuntimeError
#             #     # "You cannot use AsyncToSync in the same thread as an async event loop - "
#             #     # "just await the async function directly."
#             #     value = func(*args, **kwargs)
#         else:
#             # If running async, call async func normally
#             value = func(*args, **kwargs)
#         return value
#     return decorator



import os
import shutil
import uvicore
import fileinput
from uvicore import log
from typing import List
from uvicore.support import str
from uvicore.support.dumper import dump, dd


class Schematic:

    def __init__(self, *, type: str, stub: str, dest: str, replace: List):
        self.type = type
        self.package = uvicore.app.package(main=True)
        self.stub = os.path.realpath(stub)
        self.dest = os.path.realpath(dest)
        self.path = "/".join(self.dest.split('/')[0:-1])

        # Replacements (order is important)
        self.replacements = replace
        self.replacements.extend([
            ("xx_vendor", self.package.vendor),
            ("xx_Vendor", str.studly(self.package.vendor)),
            ("xx_appname", self.package.short_name),
            ("xx_AppName", str.studly(self.package.short_name)),
        ])

    def generate(self):
        log.header('Generating new schematic')
        log.item2('Type: ' + self.type)
        log.item2('Stub: ' + self.stub)
        log.item2('Destination: ' + self.dest)
        log.item2('Replacements:')
        log.dump(self.replacements)

        if not os.path.exists(self.path):
            os.mkdir(self.path)

        if os.path.exists(self.dest):
            #raise Exception('Destination file already exists')
            log.nl(); log.error('ERROR: Destination file already exists.  Schematic cancelled.')
            exit()

        # Copy stub
        log.item4('Copying stub to destination')
        shutil.copyfile(self.stub, self.dest)

        # Search and Replace
        log.item4('Searching and replacing keywords in destination stub')
        with fileinput.FileInput(self.dest, inplace=True) as f:
            for line in f:
                for replacement in self.replacements:
                    line = line.replace(replacement[0], replacement[1])
                print(line, end="")





import uvicore
import sqlalchemy as sa
from uvicore.database import Table
from uvicore.support.dumper import dump, dd


# Get related tablenames with proper prefixes
users = uvicore.db.tablename('auth.users')
groups = uvicore.db.tablename('auth.groups')

@uvicore.table()
class UserGroups(Table):

    # Actual database table name
    # Plural table names and singluar model names are encouraged
    # Do not add a package prefix, leave that to the connection config
    name = 'user_groups'

    # Connection for this database from your config file
    connection = 'auth'

    # SQLAlchemy Table definition as a list (exclude name and metadata)
    # This will be converted into an actual SQLAlchemy Table() instance
    # See https://docs.sqlalchemy.org/en/13/core/schema.html
    schema = [
        sa.Column('user_id', sa.Integer, sa.ForeignKey(f"{users}.id"), nullable=False),
        #sa.Column('group_key', sa.String(length=20), sa.ForeignKey(f"{groups}.key"), nullable=False),
        sa.Column('group_id', sa.Integer, sa.ForeignKey(f"{groups}.id"), nullable=False),
        sa.PrimaryKeyConstraint('user_id', 'group_id')
    ]

    # Optional SQLAlchemy Table() instance kwargs
    schema_kwargs = {}



from __future__ import annotations
import uvicore
from uvicore.typing import Optional, List
from uvicore.support.dumper import dd, dump
from uvicore.auth.models.permission import Permission
from uvicore.auth.database.tables import roles as table
from uvicore.orm import Model, ModelMetaclass, Field, BelongsToMany


@uvicore.model()
class Role(Model['Role'], metaclass=ModelMetaclass):
    """Auth Role Model"""

    __tableclass__ = table.Roles

    id: Optional[int] = Field('id',
        primary=True,
        description='Role ID',
    )

    name: str = Field('name',
        description='Role Name',
    )

    # superadmin: bool = Field('superadmin',
    #     default=False,
    #     description="Role is SuperAdmin"
    # )

    # Many-To-Many via role_permissions pivot table
    permissions: Optional[List[Permission]] = Field(None,
        description="Role Permissions",
        relation=BelongsToMany('uvicore.auth.models.permission.Permission', join_tablename='role_permissions', left_key='role_id', right_key='permission_id'),
    )



import sys
import uvicore
import logging
from uvicore.support.printer import prettyprinter


def dump(*args):
    """Dump variables using prettyprinter"""

    # Detect if running in pytest
    if "pytest" in sys.modules: level = None

    for arg in args:
        if type(arg) == str:
            # I don't want string printed with dump because it adds quotes to the string
            # which seems confusing at times.
            #prettyprinter.cpprint(arg, width=10000, ribbon_width=10000)
            print(arg)
        else:
            width = 120
            if uvicore.config:
                if uvicore.config.app.dump.width:
                    width = uvicore.config.app.dump.width
            prettyprinter.cpprint(arg, width=width, ribbon_width=width)


def log_dump(*args):
    """Dump variables using prettyprinter and obey log levels.

    Only dump if log level INFO or DEBUG.
    Used from uvicore logger log.dump()"""

    # Get current log level of Console logger
    # Uvicore may not be loaded yet, if not, set default level to INFO
    level = None
    if uvicore.log: level = logging.getLevelName(uvicore.log.console_handler.level)

    # Detect if running in pytest
    if "pytest" in sys.modules: level = None

    # Only dump if Console log level is DEBUG or INFO
    if level == 'INFO' or level == 'DEBUG' or level is None:
        dump(*args)


def dd(*args):
    """Dump variables using prettyprinter and exit()"""
    dump(*args)
    exit()



import uvicore
import sqlalchemy as sa
from uvicore.database import Table
from uvicore.support.dumper import dump, dd


@uvicore.table()
class Groups(Table):

    # Actual database table name
    # Plural table names and singluar model names are encouraged
    # Do not add a package prefix, leave that to the connection config
    name = 'groups'

    # Connection for this database from your config file
    connection = 'auth'

    # SQLAlchemy Table definition as a list (exclude name and metadata)
    # This will be converted into an actual SQLAlchemy Table() instance
    # See https://docs.sqlalchemy.org/en/13/core/schema.html
    schema = [
        sa.Column('id', sa.Integer, primary_key=True),
        #sa.Column('key', sa.String(length=20), primary_key=True),
        sa.Column("name", sa.String(length=100))
    ]

    # Optional SQLAlchemy Table() instance kwargs
    schema_kwargs = {}



import uvicore
from uvicore.support.dumper import dump, dd
from uvicore.support.module import location, load
from uvicore.typing import List, Dict, OrderedDict, Union
from uvicore.contracts import Provider
from uvicore.support.collection import unique



class Http(Provider):
    """Http Service Provider Mixin"""

    def register_http_web_routes(self, module: str, prefix: str, name_prefix: str = ''):
        # Default registration
        self.package.registers.defaults({'web_routes': True})

        # We do NOT check if package can registers.web_routes here as we always want
        # the package definition to contain the routes.  Instead we simply don't load
        # the routes in http/bootstrap.py merge_routes() if registers is False

        # Add routes to package definition
        self.package.web.routes_module = module
        self.package.web.prefix = prefix
        self.package.web.name_prefix = name_prefix

    def register_http_api_routes(self, module: str, prefix: str, name_prefix: str = 'api'):
        # Default registration
        self.package.registers.defaults({'api_routes': True})

        # We do NOT check if package can registers.api_routes here as we always want
        # the package definition to contain the routes.  Instead we simply don't load
        # the routes in http/bootstrap.py merge_routes() if registers is False

        # Add routes to package definition
        self.package.api.routes_module = module
        self.package.api.prefix = prefix
        self.package.api.name_prefix = name_prefix

    # def middlewareNO(self, items: OrderedDict):
    #     # Don't think this should be here.  middleware is app global
    #     # not per package.  I add apps middleware in http service.py
    #     # Default registration
    #     self.package.registers.defaults({'middleware': True})

    #     # Register middleware only if allowed
    #     #if self.package.registers.middleware:
    #         #self._add_http_definition('middleware', items)

    def register_http_views(self, modules: List):
        # Default registration
        self.package.registers.defaults({'views': True})

        # Register views only if allowed
        if self.package.registers.views:
            self.package.web.view_paths = modules

    def register_http_view_context_processors(self, items: Dict):
        # Default registration - template obeys view registration
        self.package.registers.defaults({'views': True})

        # Register templates only if [views] allowed
        if self.package.registers.views:
            self.package.web.context_processors = Dict(items)

    def register_http_view_composers(self, module: Union[str, Dict], views: Union[str, List] = None, *, append: bool = False):
        # Default registration - we use same 'views' registration for composers as well
        self.package.registers.defaults({'views': True})

        # Register composers only if allowed
        if self.package.registers.views:
            if type(module) == dict:
                # Dict based, example:
                # self.composers({
                #     'mreschke.themes.http.composers.theme.Theme': 'themes/*',
                #     'mreschke.themes.http.composers.theme.Theme2': ['themes/*', 'themes2/*'],
                # })
                for (mod, views) in module.items():
                    self.composers(mod, views)
            else:
                # Parameter based
                if views is None:
                    # Delete this entire view composer
                    del self.package.web.view_composers[module]
                    return
                if type(views) != list: views = [views]
                if append:
                    # Append views to this modules List
                    if not self.package.web.view_composers[module]: self.package.web.view_composers[module] = []
                    self.package.web.view_composers[module].extend(views)
                    self.package.web.view_composers[module] = unique(self.package.web.view_composers[module])
                else:
                    self.package.web.view_composers[module] = views


    def register_http_assets(self, items: List):
        # Default registration
        self.package.registers.defaults({'assets': True})

        # Register assets only if allowed
        if self.package.registers.assets:
            self.package.web.asset_paths = items

    def register_http_public(self, items: List):
        self.package.web.public_paths = items





from __future__ import annotations
from typing import Optional, List
from uvicore.contracts import Model as ModelInterface


# Model interfaces, though redundant, are used for proper type hinting code intellisense
class UserInfo(ModelInterface):
    id: Optional[int]
    extra1: str
    user_id: int
    user: Optional[User]


# Import relations after model to avoid circular dependencies
from .user import User







# @uvicore.service()
# class Routes(RoutesInterface, Generic[R]):

#     endpoints: str = None

#     @property
#     def app(self) -> ApplicationInterface:
#         return self._app

#     @property
#     def package(self) -> PackageInterface:
#         return self._package

#     @property
#     def Router(self) -> R:
#         return self._Router

#     @property
#     def prefix(self) -> str:
#         return self._prefix

#     def __init__(self,
#         app: ApplicationInterface,
#         package: PackageInterface,
#         Router: R,
#         prefix: str
#     ):
#         self._app = app
#         self._package = package
#         self._Router = Router
#         self._prefix = prefix


#     def new_router(self):
#         router = self.Router()

#         # Add route context into Router
#         router.uvicore = Dict({
#             'prefix': self.prefix,
#             'endpoints': self.endpoints,
#         })
#         return router


#     def include(self, module, *, prefix: str = '', tags: List[str] = None) -> None:
#         #self.http.controller(controller.route, prefix=self.prefix)

#         if type(module) == str:
#             # Using a string to point to an endpoint class controller
#             controller = load(self.endpoints + '.' + module + '.route')
#             uvicore.app.http.include_router(
#                 controller.object,
#                 prefix=self.prefix + str(prefix),
#                 tags=tags,
#             )
#         else:
#             # Passing in an actual router class
#             uvicore.app.http.include_router(
#                 module,
#                 prefix=self.prefix + str(prefix),
#                 tags=tags,
#             )

#     # def Router(self) -> R:
#     #     return self._Router()


# # IoC Class Instance
# #Routes: RoutesInterface = uvicore.ioc.make('Routes', _Routes)

# # Public API for import * and doc gens
# #__all__ = ['Routes', '_Routes']



import uvicore
from uvicore.typing import List, Union
from uvicore.contracts import UserInfo as UserInfoInterface
from uvicore.support.dumper import dump, dd
from pydantic import BaseModel as PydanticModel


# Pydantic model because we often return this class in API endpoints
# Endpoints can return classes, but pydantic forces a __initialized__ property
# on them which is annoying.
@uvicore.service()
class UserInfo(PydanticModel, UserInfoInterface):
    """Auth logged in user dataclass representation

    Not to be confused with the User database ORM model.  Logged in Auth User
    requires its own interface and abstraction from any model.  We call it UserInfo.
    """

    # Required even through they are in the contract simply because
    # this is a pydantic model
    id: int
    uuid: str
    sub: str
    username: str
    email: str
    first_name: str
    last_name: str
    title: str
    avatar: str
    groups: List[str]
    roles: List[str]
    permissions: List[str]
    superadmin: bool
    authenticated: bool

    @property
    def name(self):
        """First and last name"""
        return str(self.first_name + ' ' + self.last_name).strip()

    @property
    def avatar_url(self):
        """Alias to avatar"""
        return self.avatar

    @property
    def admin(self):
        """Check if user is a superadmin"""
        return self.superadmin

    @property
    def is_admin(self):
        """Check if user is a superadmin"""
        return self.superadmin

    @property
    def is_superadmin(self):
        """Check if user is a superadmin"""
        return self.superadmin

    @property
    def is_not_admin(self):
        """Check if user is not a superadmin"""
        return not self.superadmin

    @property
    def is_authenticated(self):
        """Check if user is logged in"""
        return self.authenticated

    @property
    def loggedin(self):
        """Check if user is logged in"""
        return self.authenticated

    @property
    def is_loggedin(self):
        """Check if user is logged in"""
        return self.authenticated

    @property
    def is_not_loggedin(self):
        """Check if user is not logged in"""
        return not self.authenticated

    @property
    def is_not_authenticated(self):
        """Check if user is not logged in"""
        return not self.authenticated

    @property
    def check(self):
        """Alias to authenticated"""
        return self.authenticated

    def can(self, permissions: Union[str, List]) -> bool:
        """Check if user has ALL of these permissions"""
        if self.superadmin: return True

        permissions = [permissions] if isinstance(permissions, str) else list(permissions)

        for permission in permissions:
            if permission not in self.permissions:
                return False
        return True

    def can_any(self, permissions: Union[str, List]) -> bool:
        """Check if user has any one of these permissions"""
        if self.superadmin: return True

        permissions = [permissions] if isinstance(permissions, str) else list(permissions)

        for permission in permissions:
            if permission in self.permissions:
                return True
        return False

    def cant(self, permissions: Union[str, List]) -> bool:
        """Check if user does not have one of these permissions"""
        return not self.can(permissions)

    def cannot(self, permissions: Union[str, List]) -> bool:
        """Alias to cant"""
        return self.cant(permissions)






from __future__ import annotations
from typing import Optional, List
from uvicore.contracts import Model as ModelInterface


# Model interfaces, though redundant, are used for proper type hinting code intellisense
class User(ModelInterface):
    id: Optional[int]
    email: str
    info: Optional[UserInfo]


# Import relations after model to avoid circular dependencies
from .user_info import UserInfo



import uvicore
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Logger as LoggerInterface

class Logger:

    def make(self, Logger: LoggerInterface):
        config = uvicore.config('app.logger')
        return Logger(config)



import uvicore
import sqlalchemy as sa
from abc import ABCMeta
from typing import Dict, List
from uvicore.support.dumper import dd, dump

@uvicore.service()
class Table:

    @property
    def table(self) -> sa.Table:
        return self.schema

    def __init__(self):
        # Tables are singleton classes bound in the IoC
        # So they are instantiated ONCE when first ioc.make()
        # Once instantiated, the actual SQLAlchemy table is created
        # and added to the metadata.  Tables cannot be made twice or SA
        # complains about duplicate tables.  To override a table simply use
        # your app configs bindings array to swap the initial singleton binding.
        self.metadata = uvicore.db.metadata(self.connection)
        prefix = uvicore.db.connection(self.connection).prefix
        if prefix is not None:
            self.name = str(prefix) + self.name

        # Only enhance schema if connection string backend is 'sqlalchemy'
        if uvicore.db.connection(self.connection).backend == 'sqlalchemy':
            self.schema: sa.Table = sa.Table(
                self.name,
                self.metadata,
                *self.schema,
                **self.schema_kwargs
            )


# class SchemaOLD(ABCMeta):
#     def __new__(mcs, name, bases, namespace, **kwargs):
#         # Set metadata based on connection string
#         namespace['metadata'] = uvicore.db.metadata(namespace.get('connection'))

#         # Add prefix to table name
#         prefix = uvicore.db.connection(namespace.get('connection')).prefix
#         if prefix is not None:
#             namespace['name'] = str(prefix) + namespace['name']

#         # Convert table List into actual SQLAlchemy table schema
#         namespace['schema'] = sa.Table(
#             namespace['name'],
#             namespace['metadata'],
#             *namespace['schema'],
#             **namespace.get('schema_kwargs')
#         )

#         # Return this new metaclass
#         return super().__new__(mcs, name, bases, namespace, **kwargs)

# class Table:
#     def __init__(self,
#         name: str,
#         connection: str,
#         schema: sa.Table,
#      ) -> None:
#         self.name = name
#         self.connection = connection
#         self.schema = schema

# autocolumns = [
#     sa.Column('first_name', sa.String(length=50))
# ]


# class Table(BaseTable):
#     """A simple SQLAlchemy Table abstraction that allows you to specify
#     a connection string instead of metedata.  Metadata is derived from
#     this connection string.  Connection string now available at yourtable.connection
#     """
#     def __new__(cls, name: str, connection: str, *args, **kwargs):
#         return super().__new__(cls, name, db.metadata.get(connection), *args, *kwargs)

#     def __init__(self, name: str, connection: str, *args, **kwargs) -> None:
#         # Don't set self.name, it already exists in BaseTable
#         # We are only adding a connection property to the already large
#         # BaseTable dictionary
#         self.connection = connection
#         super().__init__(self, name, db.metadata.get(connection), *args, **kwargs)



import uvicore
from typing import Dict
from uvicore.package import Provider
from uvicore.support.dumper import dump, dd
from uvicore.console.package.registers import Cli


@uvicore.provider()
class Orm(Provider, Cli):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # ???
        # Maybe provide a self.event_strings({}) to at least define which event strings there are
        # then you can show in ./uvicore event list

        # Register events used in this package
        # self.events.register(
        #     name='uvicore.orm-{dynamic.your.model}-BeforeSave',
        #     description='Dynamic, before saving (insert or update) model to db.',
        #     dynamic=True,
        #     is_async=True
        # )
        # self.events.register(
        #     name='uvicore.orm-{your.model}-AfterSave',
        #     description='Dynamic, after saving (insert or update) model to db.',
        #     dynamic=True,
        #     is_async=True,
        # )

        # {model} is would be full app1.models.post.Post
        # uvicore.orm-app1.models.post.Post-Saving

        # Listen to all ORM saving
        #orm-*-saving

        # Listen to all post model events
        #orm-{post}-*

        # def test(event, payload):
        #     print('hi')

        #dd(uvicore.events.event('uvicore.foundation.events.app.Registered'))

        # Register IoC bindings
        # Automatic - self.bind('Model', 'uvicore.orm.model._Model', aliases=['model'])
        # Automatic - self.bind('ModelMetaclass', 'uvicore.orm.metaclass._ModelMetaclass')

        #self.bind('Model', 'uvicore.orm.model._Model', aliases=['model'])
        #self.bind('ModelMetaclass', 'uvicore.orm.metaclass._ModelMetaclass')

    def boot(self) -> None:
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define commands
        self.register_cli_commands({
            # Extend schematic generator commands
            'gen': {
                'commands': {
                    'model': 'uvicore.orm.commands.generators.model',
                },
            }
        })



import os
from subprocess import Popen

import uvicorn

import uvicore
from uvicore.console import command


@command()
def cli():
    """Unicorn dev server (reload and logs)"""
    # Uvicorn dev server info
    appname = uvicore.config('app.server.app')
    host = uvicore.config('app.server.host')
    port = uvicore.config('app.server.port')
    autoreload = uvicore.config('app.server.reload')
    access_log = uvicore.config('app.server.access_log')

    # Run Uvicorn server
    uvicorn.run(appname, host=host, port=port, reload=autoreload, access_log=access_log)



import uvicore
from uvicore.events import Event


@uvicore.event()
class Startup(Event):
    """HTTP Server has been started.  This is the Starlette startup async event."""

    is_async = True

    def __init__(self):

        pass


@uvicore.event()
class Shutdown(Event):
    """HTTP Server has been shutdown.  This is the Starlette shutdown async event."""

    is_async = True

    def __init__(self):
        self.is_async = True
        pass





# # Register events used in this package
# self.events.register(
#     name='uvicore.http.events.server.Startup',
#     description='HTTP Server has been started.  This is the Starlette startup event.',
#     is_async=False,
# )
# self.events.register(
#     name='uvicore.http.events.server.Shutdown',
#     description='HTTP Server has been shutdown.  This is the Starlette shutdown event.',
#     is_async=False,
# )



import uvicore
from uvicore.package import Provider
from uvicore.support.dumper import dump, dd
from uvicore.console.package.registers import Cli


@uvicore.provider()
class Foundation(Provider, Cli):

    def register(self):
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # Register configs
        self.configs([
            {'key': self.name, 'module': 'uvicore.foundation.config.package.config'}
        ])

        # Register events used in this package
        # self.events.register(
        #     name='uvicore.foundation.events.app.Registered',
        #     description='Application bootstrap has registered all package service providers.',
        #     dynamic=False,
        #     is_async=False,
        # )
        # self.events.register(
        #     name='uvicore.foundation.events.app.Booted',
        #     description='Application bootstrap has booted all package service providers.',
        #     dynamic=False,
        #     is_async=False,
        # )

        # # Register root level commands
        # self.commands(
        #     name='root',
        #     #help='xx',
        #     commands=[
        #         ('version', 'uvicore.foundation.commands.version.cli'),
        #     ]
        # )

    def boot(self):
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        # Define service provider registration control
        # No - Never allow this packages registrations to be disabled from other configs

        # Define commands
        # Why here in foundation?  Because these items are not actual packages
        # with their own service providers to register themselves.
        # self.commands({
        #     # Register Ioc commands
        #     'ioc': {
        #         'help': 'Uvicore Ioc (Inversion of Control) Information',
        #         'commands': {
        #             'bindings': 'uvicore.container.commands.ioc.bindings',
        #             'singletons': 'uvicore.container.commands.ioc.singletons',
        #             'overrides': 'uvicore.container.commands.ioc.overrides',
        #             'type': 'uvicore.container.commands.ioc.type',
        #             'get': 'uvicore.container.commands.ioc.get',
        #         },
        #     },

        #     # Register Package commands
        #     'package': {
        #         'help': 'Uvicore Package Information',
        #         'commands': {
        #             'providers': 'uvicore.package.commands.package.providers',
        #             'list': 'uvicore.package.commands.package.list',
        #             'get': 'uvicore.package.commands.package.get',
        #         },
        #     },

        #     # Register Event commands
        #     'event': {
        #         'help': 'Uvicore Event Information',
        #         'commands': {
        #             'list': 'uvicore.events.commands.event.list',
        #             'get': 'uvicore.events.commands.event.get',
        #             'listeners': 'uvicore.events.commands.event.listeners',
        #         },
        #     },
        # })

        # Alternative in kwargs format

        # Register App commands
        self.register_cli_commands(
            group='app',
            help='Uvicore Application Information',
            commands={
                'info': 'uvicore.foundation.commands.app.info',
            }
        )


        # Register Ioc commands
        self.register_cli_commands(
            group='ioc',
            help='Uvicore Ioc (Inversion of Control) Information',
            commands={
                'bindings': 'uvicore.container.commands.ioc.bindings',
                'singletons': 'uvicore.container.commands.ioc.singletons',
                'overrides': 'uvicore.container.commands.ioc.overrides',
                'type': 'uvicore.container.commands.ioc.type',
                'get': 'uvicore.container.commands.ioc.get',
            }
        )

        # Register Package commands
        self.register_cli_commands(
            group='package',
            help='Uvicore Package Information',
            commands={
                'providers': 'uvicore.package.commands.package.providers',
                'list': 'uvicore.package.commands.package.list',
                'get': 'uvicore.package.commands.package.get',
            }
        )

        # Register Event commands
        self.register_cli_commands(
            group='event',
            help='Uvicore Event Information',
            commands={
                'list': 'uvicore.events.commands.event.list',
                'get': 'uvicore.events.commands.event.get',
                'listeners': 'uvicore.events.commands.event.listeners',
            },
        )

        # Alternative, pull from a config
        #self.commands(self.package.config.commands)



# type: ignore
from .event import Event
from .dispatcher import Dispatcher
from .handler import Handler

# The events package uses a __init__.py because users will import these
# methods from their own class based events and we want a nicer import which
# looks like - from uvicore.events import Event
